from copy import deepcopy
from xml.dom import NotSupportedErr

from numpy import (
    arctan2,
    asarray,
    concatenate,
    cos,
    diag,
    dot,
    max,
    mean,
    min,
    sin,
    sqrt,
    square,
)
from numpy import sum as nsum
from numpy import tile, zeros
from numpy.linalg import inv, pinv

from pydistsim.logging import logger
from pydistsim.utils.localization.aoastitcher import AoAStitcher
from pydistsim.utils.localization.diststitcher import DistStitcher
from pydistsim.utils.memory.positions import Positions


def align_clusters(dst, src, scale):
    """Scale, rotate, translate srcLoc w.r.t. dst."""
    assert isinstance(src, Positions)
    assert isinstance(dst, Positions)
    if scale:
        stitcher = AoAStitcher(reflectable=True)
    else:
        stitcher = DistStitcher()
    stitcher.align(dst, src)


def get_rms(truePos, estimated, align=False, scale=False, norm=False):
    """
    Returns root mean square error of estimated positions.

    Set align if estimated positions need to be transformed (rotated and
    translated) before calculating rms error of localization. Set scale=True
    if they needs to be scaled also.

    If norm is True then rms is divided with norm of the true formation
    translated to centroid.

    """
    truePos = Positions.create(truePos)
    assert len(truePos) == 1
    estimated = Positions.create(estimated)

    if align:
        estimated = deepcopy(estimated)
        align_clusters(truePos, estimated, scale)

    suma = 0
    node_count = 0
    for estimated_subcluster in estimated:
        for n in estimated_subcluster:
            suma += (estimated_subcluster[n][0] - truePos[0][n][0]) ** 2 + (
                estimated_subcluster[n][1] - truePos[0][n][1]
            ) ** 2
            node_count += 1

    rms = sqrt(suma / node_count)
    if norm:
        sc = truePos.subclusters[0]
        truePos.subclusters[0] = {n: p for n, p in list(sc.items()) if n in estimated.subclusters[0]}
        rms = rms / get_pos_norm(truePos)
    return rms


def construct_G(pos, edges, u, sensor):
    """
    Construct Jacobian of vector mu(pos) that defines expectation of
    measurement given  positions.

    u - number of unknowns i.e. 3 - (x, y, theta)
    edges - list of edges as tuples, if there is (i, j) then there should not
            be (j, i)
    sensor - 'DistSensor' or 'AoASensor'

    For AoA and measurement between nodes i and j:
    mu(pos) = arctan((pos_y_j - pos_y_i)/(pos_x_j - pos_x_i) - alpha_i)

    """
    nodes = list(pos.keys())
    neighbors = {n: [] for n in nodes}
    for n1, n2 in edges:
        neighbors[n1].append(n2)
        neighbors[n2].append(n1)
    # number of measurements x number of unknowns
    G = zeros((2 * len(edges), u * len(nodes)))
    m = 0
    for r, node in enumerate(nodes):
        (x_r, y_r) = pos[node][:2]
        for neighbor in neighbors[node]:
            t = nodes.index(neighbor)
            (x_t, y_t) = pos[neighbor][:2]
            d = sqrt((x_r - x_t) ** 2 + (y_r - y_t) ** 2)
            if sensor == "DistSensor":
                G[m, r * u] = (x_r - x_t) / d
                G[m, r * u + 1] = (y_r - y_t) / d
                G[m, t * u] = (x_t - x_r) / d
                G[m, t * u + 1] = (y_t - y_r) / d
            elif sensor == "AoASensor":
                G[m, r * u] = (y_t - y_r) / d**2
                G[m, r * u + 1] = (x_r - x_t) / d**2
                G[m, t * u] = (y_r - y_t) / d**2
                G[m, t * u + 1] = (x_t - x_r) / d**2
                if u == 3:
                    G[m, t * u + 2] = -1.0
            m += 1  # next measurement
    return G


# TODO: only one per row -1 element in G?
def get_crb(net, sensor, compass="off", loc_type="anchor-free", anchors=[]):
    """
    Calculates Cramer-Rao lower bound on covariance matrix of unbiased multihop
    location estimator based on sensor (distance/angle) measurement between
    in-commrange neighbors.

    POS is n by 2 matrix of X and Y coordinates of n nodes
    SIGMA is standard deviation on measurement vector.

    Arguments:
     `compass` -- 'off' (default) or 'on' which means all nodes have compass
     `loc_type` -- 'anchor-free' (default) or 'anchored'
     `anchors` -- list of anchor nodes, used only when loc_type is 'anchored'

    References:
        Savvides2003 - Savvides et al,
                        On the Error Characteristics of Multihop Node
                        Localization in Ad-Hoc Sensor Networks, 2003
        Patwari2005 - Patwari et al,
                        Locating the nodes: cooperative localization in
                        wireless sensor networks
        Chang2006 - Chen Chang et al,
                        Cramer-Rao-Type Bounds for Localization, 2006
    """

    if loc_type == "anchored" and not anchors:
        raise Exception("Anchors not defined.")
    if loc_type == "anchor-free" and anchors:
        logger.warning("Anchors are ignored")
        anchors = []

    nodes = [node for node in net.nodes() if node not in anchors]

    if sensor.name() == "AoASensor" and compass == "off":
        u = 3  # x, y, theta
    else:  # DistSensor or AoASensor with compass on
        u = 2  # x, y

    G = construct_G(net.pos, net.edges(), u, sensor.name())

    for s in nodes[0].compositeSensor.sensors:
        if s.name() == sensor.name():
            sigma = s.probabilityFunction.scale
            break
    else:
        raise Exception("Sensor not found in nodes")

    J = (dot(G.T, G)) / sigma**2

    if loc_type == "anchor-free":
        cov = pinv(J)
    elif loc_type == "anchored":
        cov = inv(J)

    # Chang2006 (28)
    di = diag(cov)
    # extract only position from cov
    di = concatenate((di[::3], di[1::3]))

    # return is lower bound on rms -> sqrt(1/n*sum_i^n((x_i'-x_i)^2+(y_i'-y_i)^2))
    return sqrt(2 * mean(di))


def get_crb_norm(*args, **kwargs):
    """
    Calculates crb that is not dependent on scale i.e. to check correlation
    with gdop.
    """
    net = args[0]
    norm = get_pos_norm(net.pos)
    return get_crb(*args, **kwargs) / norm


def get_pos_norm(pos):
    """Translate positions so that centroid is in the origin and return mean
    norm of the translated positions."""
    pos = Positions.create(pos)
    assert len(pos) == 1
    n = len(pos[0])
    p = zeros((n, 2))
    for i, node in enumerate(pos[0]):
        p[i, :] = pos[0][node][:2]
    centroid = p.sum(axis=0) / n
    p -= tile(centroid, (n, 1))
    p_norm = nsum(sqrt(nsum(square(p), axis=1))) / n
    return p_norm


def get_aoa_gdop_rel(estimated):
    """
    Calculation of relative GDOP is based on CRB.

    GDOP_rel = sigma_CRB/sigma_d = sqrt(tr((G^TG)^-1)/(sum(Di^2))/M)
    As regular GDOP, it is not dependent on scale nor on sigma_AoA.
    """
    estimated = Positions.create(estimated)
    assert len(estimated.subclusters) == 1
    pos = estimated.subclusters[0]
    nodes = list(pos.keys())
    edges = [e for e in list(pos.keys())[0].network.edges() if e[0] in nodes and e[1] in nodes]
    G = construct_G(pos, edges, 3, "AoASensor")
    J = dot(G.T, G)
    cov = pinv(J)
    di = diag(cov)
    di = concatenate((di[::3], di[1::3]))
    var_p = sum(di) / len(nodes)
    distances = [sqrt(dot(pos[n1][:2] - pos[n2][:2], pos[n1][:2] - pos[n2][:2])) for n1, n2 in edges]
    var_d = sum(square(distances)) / len(edges)
    return sqrt(var_p / var_d)


def get_aoa_gdop_node(estimated, node):
    """
    Calculate geometric dilution of precision for node in estimated formation.
    Node is in the formation and all other nodes should be neighbors of
    node. Node should have Neighbors sensorReadings in it memory.
    """
    if "AoASensor" not in [sensor.name() for sensor in node.sensors]:
        raise NotSupportedErr("Only angle of arrival based gdop is supported")
    sensor = node.compositeSensor.get_sensor("AoASensor")
    sigma = sensor.probabilityFunction.scale
    for n in estimated:
        sensor = n.compositeSensor.get_sensor("AoASensor")
        if sensor.probabilityFunction.scale != sigma:
            raise NotSupportedErr("All nodes' AoA sensors should have same scale")
    if len(estimated) <= 2:
        return 0.0
    # Note from Torrieri, Statistical Theory of Passive Location Systems
    # if measurement sigmas are all equal gdop doesn't depend on sigma.
    sigma = 1

    x, y = estimated[node][:2]
    fi = []
    d = []

    for n in estimated:
        xi, yi = estimated[n][:2]
        if n != node and n in node.memory["sensorReadings"]["Neighbors"]:
            fi.append(arctan2(y - yi, x - xi))
            d.append(sqrt((x - xi) ** 2 + (y - yi) ** 2))

    mi = ni = l = 0
    for fii, di in zip(fi, d):
        mi += (cos(fii) / di / sigma) ** 2  # (129)
        l += (sin(fii) / di / sigma) ** 2  # (130)
        ni += sin(fii) * cos(fii) / (di * sigma) ** 2  # (131)

    sigma1 = sqrt(mi / (mi * l - ni**2))  # (126)
    sigma2 = sqrt(l / (mi * l - ni**2))  # (127)
    # sigma12 = sqrt(ni/(mi*l-ni**2))

    sigmad = sqrt(sum((di * sigma) ** 2 for di in d) / len(d))  # (138)
    return sqrt(sigma1**2 + sigma2**2) / sigmad  # (139)


def get_aoa_gdops(estimated):
    estimated = Positions.create(estimated)
    assert len(estimated.subclusters) == 1
    estimated = estimated.subclusters[0]
    return [get_aoa_gdop_node(estimated, node) for node in estimated]


def get_aoa_gdop(estimated):
    return sum(get_aoa_gdops(estimated))


def show_subclusters(net, subclusters):
    """Show colored subclusters."""
    colors = [node in subclusters[0] and "g" or "r" for node in net.nodes()]
    net.show(nodeColor=colors)


def show_localized(net, estimated, scale=False, align=True, display_loc_err=True, show_labels=True):
    """
    Display estimated positions.

    estimated should be a list of dictionaries.

    """
    from matplotlib.collections import LineCollection
    from matplotlib.pylab import gca

    truePos = Positions.create(net.pos)
    estimated = Positions.create(estimated)
    # copy estimated so that passed estimated remains unchanged
    estimated = deepcopy(estimated)
    if align:
        # rotate, translate and if needed scale estimated w.r.t. true positions
        align_clusters(truePos, estimated, scale)

    # TODO: implement display of all subclusters
    if len(estimated) > 1:
        raise (NotImplementedError)
    else:
        estimated_sc = estimated[0]

        # net.show(positions=estimated_sc, show_labels=show_labels)
        fig = net.get_fig(positions=estimated_sc, show_labels=show_labels)
        ax = fig.gca()
        minpos = min(list(estimated_sc.values()), axis=0)
        maxpos = max(list(estimated_sc.values()), axis=0)
        minpos -= (maxpos - minpos) * 0.1
        maxpos += (maxpos - minpos) * 0.1

        ax.set_xlim(minpos[0], maxpos[0])
        ax.set_ylim(minpos[1], maxpos[1])
        # fig.show()
        if display_loc_err:
            # TODO: not working in ipython notepad
            ax = gca()
            ax.set_title("Localized positions")
            ax.set_title("Localization error display")
            edge_pos = asarray([(net.pos[n], estimated_sc[n]) for n in list(estimated_sc.keys())])
            errorCollection = LineCollection(edge_pos, colors="r", transOffset=ax.transData)
            errorCollection.set_zorder(1)  # errors go behind nodes
            ax.add_collection(errorCollection)
            ax.figure.canvas.draw()



from pydistsim.demo_algorithms.broadcast import Flood
from pydistsim.network import NetworkGenerator
from pydistsim.simulation import Simulation
from pydistsim.utils.testing import PyDistSimTestCase

HELLO = "Hello distributed world"
BYE = "Bye bye distributed world"


class TestBroadcastSimple(PyDistSimTestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100, directed=False)
        self.net = net_gen.generate_random_network()
        self.sim = Simulation(self.net)
        self.sim.algorithms = ((Flood, {"informationKey": "greet", "initial_information": HELLO}),)

    def test_broadcast(self):
        sim = self.sim
        algo = sim.algorithms[0]

        for node in self.net.nodes():
            assert "greet" not in node.memory

        sim.run(1)

        algo.check_algorithm_initialization()

        sim.run(100_000)

        algo.check_algorithm_termination()

        for node in self.net.nodes():
            self.assertEqual(node.memory["greet"], HELLO)


class TestBroadcastConcatenated(PyDistSimTestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100, directed=False)
        self.net = net_gen.generate_random_network()
        self.sim = Simulation(self.net)
        self.sim.algorithms = (
            (Flood, {"informationKey": "greet", "initial_information": HELLO}),
            (Flood, {"informationKey": "bye", "initial_information": BYE}),
        )
        # Asigna el mensaje a enviar, la información inicial
        self.initiator = self.net.nodes_sorted()[0]

    def test_broadcast(self):
        sim = self.sim
        first_algo = sim.algorithms[0]
        last_algo = sim.algorithms[-1]

        for node in self.net.nodes():
            with self.subTest(node=node):
                assert "greet" not in node.memory
                assert "bye" not in node.memory

        sim.run(1)

        first_algo.check_algorithm_initialization()

        sim.run(100_000)

        last_algo.check_algorithm_termination()

        for node in self.net.nodes():
            with self.subTest(node=node):
                self.assertEqual(node.memory["greet"], HELLO)
                self.assertEqual(node.memory["bye"], BYE)



from numpy import arctan2, array, cos, dot, sin, sqrt
from numpy.linalg import det

from pydistsim.utils.localization.basestitcher import BaseStitcher
from pydistsim.utils.localization.stitchsubclusterselectors import (
    MaxCommonNodeSelector,
    StitchSubclusterSelectorBase,
)


class AoAStitcher(BaseStitcher):
    """
    Base class implementing subcluster stitching methods based on degrees of
    freedom given by localization using angle of arrival measurements i.e.
    translation, rotation/reflection and scaling.

    Reliable stitch with R, s and t that are not None is based on two or more
    common nodes between them.

    If rotation matrix includes reflection it is not reliable so return Nones.

    """

    def __new__(cls, *args, **kwargs):
        """Legacy: by default returns AoAStitcherHorn instance."""
        if cls is not AoAStitcher:
            return super().__new__(cls)
        return AoAStitcherHorn()

    def __init__(self, selector=None, **kwargs):
        self.selector = selector or MaxCommonNodeSelector(cn_count_treshold=2)
        assert isinstance(self.selector, StitchSubclusterSelectorBase)

    def _get_scaling_factor(self, commonNodes, dstSubPos, srcSubPos, p_d, p_s, w_d):
        s = sqrt(
            sum([sum((dstSubPos[cn][:2] - p_d) ** 2) * w_d[cn] for cn in commonNodes])
            / sum([sum((srcSubPos[cn][:2] - p_s) ** 2) * w_d[cn] for cn in commonNodes])
        )
        return s

    def _get_rotation_matrix_2_common_nodes(self, commonNodes, dstSubPos, srcSubPos):
        """
        Calculates rotation matrix based only on two common nodes

        Warning: should not be used when local systems could be
        reflected i.e. when they are formed by distance measurements.

        """
        vector1 = dstSubPos[commonNodes[0]] - dstSubPos[commonNodes[1]]
        vector2 = srcSubPos[commonNodes[0]] - srcSubPos[commonNodes[1]]
        theta1 = arctan2(vector1[1], vector1[0])
        theta2 = arctan2(vector2[1], vector2[0])
        theta = theta1 - theta2
        # in ccw direction
        R = array([[cos(theta), -sin(theta)], [sin(theta), cos(theta)]])
        return R


class AoAStitcherHorn(AoAStitcher):
    """
    ``reflectable`` controls whether reflection is accepted when
    calculating rotation matrix.

    :param reflectable: If True, rotation matrix can reflect nodes.
    :type reflectable: bool
    """

    def __init__(self, *args, **kwargs):
        self.reflectable = kwargs.get("reflectable", False)
        self.get_rotation_matrix = self._get_rotation_matrix_horn
        super().__init__(*args, **kwargs)

    def stitch_subclusters(self, dstSubPos, srcSubPos):
        commonNodes = self._get_common_nodes(dstSubPos, srcSubPos)
        assert len(commonNodes) > 1
        (p_s, p_d, w_d) = self._get_centroids(commonNodes, dstSubPos, srcSubPos)
        s = self._get_scaling_factor(commonNodes, dstSubPos, srcSubPos, p_d, p_s, w_d)
        if len(commonNodes) == 2:
            R = self._get_rotation_matrix_2_common_nodes(commonNodes, dstSubPos, srcSubPos)
        else:
            R = self.get_rotation_matrix(commonNodes, dstSubPos, srcSubPos, p_d, p_s)
        t = p_d - dot(dot(s, R), p_s)

        # if rotation matrix reflects nodes then it is not reliable
        if det(R) < 0 and not self.reflectable:
            return (None, None, None)

        return (R, s, t)



import unittest

from pydistsim.network import CompleteRangeType, Network
from pydistsim.network.environment import Environment2D
from pydistsim.utils import tree, visualization


class TestNetwork(unittest.TestCase):
    tree_key = "T_KEY"

    def setUp(self):
        env = Environment2D()
        self.net = Network(rangeType=CompleteRangeType(env))
        self.net.environment.image[22, 22] = 0
        self.node1 = self.net.add_node(pos=[22.8, 21.8])
        self.node2 = self.net.add_node(pos=[21.9, 22.9])
        self.node3 = self.net.add_node(pos=[21.7, 21.7])

        self.node1.memory[self.tree_key] = {
            "parent": None,
            "children": [self.node2, self.node3],
        }
        self.node2.memory[self.tree_key] = {"parent": self.node1, "children": []}
        self.node3.memory[self.tree_key] = {"parent": self.node1, "children": []}

    def test_check_tree_key(self):
        """Test check_tree_key function."""
        tree.check_tree_key(self.net, self.tree_key)

    def test_get_root_node(self):
        """Test get_tree_root function."""
        root = tree.get_root_node(self.net, self.tree_key)
        assert root == self.node1, "Incorrect tree root"

    def test_not_tree_network(self):
        env = Environment2D()
        net = Network(rangeType=CompleteRangeType(env))

        with self.assertRaises(tree.TreeNetworkException):
            tree.check_tree_key(net, self.tree_key)

        net.environment.image[22, 22] = 0
        node1 = net.add_node(pos=[22.8, 21.8])
        node2 = net.add_node(pos=[21.9, 22.9])
        node3 = net.add_node(pos=[21.7, 21.7])
        with self.assertRaises(tree.Missingtree_key):
            tree.check_tree_key(net, self.tree_key)

        node1.memory[self.tree_key] = {
            "parent": None,
            "children": [node2, node3],
        }

        with self.assertRaises(tree.Missingtree_key):
            tree.check_tree_key(net, self.tree_key)

        node2.memory[self.tree_key] = {"parent": node1, "children": []}
        node3.memory[self.tree_key] = {"parent": node1, "children": []}

        tree.check_tree_key(net, self.tree_key)

    def test_visualization(self):
        """Test visualization functions."""
        visualization.show_mst(self.net, self.tree_key)

        env = Environment2D()
        net = Network(rangeType=CompleteRangeType(env))
        node1 = net.add_node(pos=[22.8, 21.8])

        node1.memory[self.tree_key] = {
            "parent": None,
            "children": [],
        }

        visualization.show_mst(net, self.tree_key)



import unittest

from pydistsim.logging import LogLevels, disable_logger, enable_logger, set_log_level


class PyDistSimTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        set_log_level(LogLevels.DEBUG)
        enable_logger()

    @classmethod
    def tearDownClass(cls):
        set_log_level(LogLevels.INFO)
        disable_logger()



from numpy import dot

from pydistsim.utils.localization.basestitcher import BaseStitcher
from pydistsim.utils.localization.stitchsubclusterselectors import (
    MaxCommonNodeSelector,
    StitchSubclusterSelectorBase,
)


class DistStitcher(BaseStitcher):
    """
    Base class implementing subcluster stitching methods based on degrees of
    freedom given by localization using distance measurements i.e.
    translation, rotation/reflection.

    For reliable stitching there should be at least three common nodes.

    """

    def __new__(cls, *args, **kwargs):
        """Legacy: by default returns DistStitcherHorn instance."""
        if cls is not DistStitcher:
            return super().__new__(cls)
        return DistStitcherHorn()

    def __init__(self, selector=None, **kwargs):
        self.selector = selector or MaxCommonNodeSelector(cn_count_treshold=3)
        assert isinstance(self.selector, StitchSubclusterSelectorBase)


class DistStitcherHorn(DistStitcher):
    def stitch_subclusters(self, dstSubPos, srcSubPos):
        commonNodes = self._get_common_nodes(dstSubPos, srcSubPos)
        assert len(commonNodes) > 2
        (p_s, p_d, w_d) = self._get_centroids(commonNodes, dstSubPos, srcSubPos)
        s = 1
        R = self._get_rotation_matrix_horn(commonNodes, dstSubPos, srcSubPos, p_d, p_s)
        # translation vector
        t = p_d - dot(R, p_s)
        return (R, s, t)



"""
:class:`MetricCollector` is an observer that collects metrics from the simulation.
"""

from time import time
from typing import TYPE_CHECKING

from pydistsim.observers import (
    AlgorithmObserver,
    NodeNetworkObserver,
    ObservableEvents,
    SimulationObserver,
)

if TYPE_CHECKING:
    from pydistsim.algorithm import BaseAlgorithm
    from pydistsim.message import Message
    from pydistsim.network import Node
    from pydistsim.simulation import Simulation


class MetricCollector(NodeNetworkObserver, AlgorithmObserver, SimulationObserver):
    """
    MetricCollector is an observer that collects metrics from the simulation.

    ### Custom metrics
    Extend this class and implement the desired event methods to collect custom metrics.
    For registering events, call the `_add_metric` method.
    Add an instance of your custom collector to the simulation observers for it to work.

    Even so, you can use the `events` attribute to register the events you want to listen to.
    This includes new custom events that you would trigger from an algorithm.

    Example of implementing a custom metric collector:

    .. code-block:: python

        class ExampleCustomMetricCollector(MetricCollector):
            events = ["example_custom_event"]

            class CustomMetricEventType(StrEnum):
                "Definition if this enum is optional. It helps to avoid typos in the event names."
                EXAMPLE_CUSTOM_EVENT_ZERO = "EXAMPLE_CUSTOM_EVENT_ZERO"
                ...

            def on_example_custom_event(self, a, b, c):
                self._add_metric(
                    self.CustomMetricEventType.EXAMPLE_CUSTOM_EVENT_ZERO,
                    {"a": a, "b": b, "c": c}
                )

    """

    def __init__(self):
        self.metrics: list[tuple[float, ObservableEvents, dict]] = []

    def _add_metric(self, event_type: "ObservableEvents", event_data: dict):
        self.metrics.append((time(), event_type, event_data))

    # AlgorithmObserver methods
    def on_step_done(self, algorithm: "BaseAlgorithm") -> None:
        self._add_metric(
            ObservableEvents.step_done,
            None,
        )

    def on_message_sent(self, message: "Message") -> None:
        self._add_metric(
            ObservableEvents.message_sent,
            {"message": message},
        )

    def on_algorithm_started(self, algorithm: "BaseAlgorithm") -> None:
        self._add_metric(
            ObservableEvents.algorithm_started,
            {"algorithm": algorithm},
        )

    # SimulationObserver methods
    def on_state_changed(self, simulation: "Simulation") -> None: ...

    def on_algorithm_finished(self, algorithm: "BaseAlgorithm") -> None:
        self._add_metric(
            ObservableEvents.algorithm_finished,
            {"algorithm": algorithm},
        )

    def on_network_changed(self, simulation: "Simulation") -> None: ...

    # NetworkObserver methods (none yet)
    def on_message_delivered(self, message: "Message") -> None:
        self._add_metric(
            ObservableEvents.message_delivered,
            {"message": message},
        )

    # NodeNetworkObserver methods
    def on_node_status_changed(self, node: "Node", previous_status: str, new_status: str) -> None:
        self._add_metric(
            ObservableEvents.node_status_changed,
            {
                "node": node,
                "previous_status": previous_status,
                "new_status": new_status,
            },
        )

    def create_report(self):
        """
        Returns a dictionary with the collected metrics.

        Currently, it returns the number of messages sent, messages delivered, and quantity of status changes.
        Future versions may include more metrics, with better reporting capabilities.
        """
        msg_sent = 0
        msg_delivered = 0
        qty_nodes_status_changed = 0
        qty_steps_done = 0

        for timestamp, event_type, event_data in self.metrics:
            match event_type:
                case ObservableEvents.message_sent:
                    msg_sent += 1
                case ObservableEvents.message_delivered:
                    msg_delivered += 1
                case ObservableEvents.node_status_changed:
                    qty_nodes_status_changed += 1
                case ObservableEvents.step_done:
                    qty_steps_done += 1

        return {
            "Qty. of messages sent": msg_sent,
            "Qty. of messages delivered": msg_delivered,
            "Qty. of status changes": qty_nodes_status_changed,
            "Qty. of steps": qty_steps_done,
        }



from collections import defaultdict
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from enum import StrEnum
from types import NoneType
from typing import TYPE_CHECKING, Union

from pydistsim.algorithm.base_algorithm import AlgorithmException, BaseAlgorithm
from pydistsim.algorithm.node_wrapper import WrapperManager
from pydistsim.logging import logger
from pydistsim.message import Message
from pydistsim.message import MetaHeader as MessageMetaHeader
from pydistsim.observers import ObservableEvents
from pydistsim.restrictions.axioms import FiniteCommunicationDelays, LocalOrientation

if TYPE_CHECKING:
    from pydistsim.algorithm.node_wrapper import NeighborLabel, NodeAccess, _NodeWrapper
    from pydistsim.network import Node


@dataclass
class Alarm:
    """
    Dataclass that represents an alarm set for a node.
    """

    time_left: int
    message: Message["_NodeWrapper"]
    node: "NodeAccess"
    triggered: bool = False


class Actions(StrEnum):
    """
    Enum that defines the possible actions that can be triggered by a message.
    """

    default = "default"
    receiving = "receiving"
    spontaneously = "spontaneously"
    alarm = "alarm"


MSG_META_HEADER_MAP = {
    MessageMetaHeader.NORMAL_MESSAGE: Actions.receiving,
    MessageMetaHeader.INITIALIZATION_MESSAGE: Actions.spontaneously,
    MessageMetaHeader.ALARM_MESSAGE: Actions.alarm,
}


ActionMethod = Callable[["NodeAlgorithm", "NodeAccess", Message], NoneType]


class StatusValues(StrEnum):
    """
    Enum that defines the possible statuses of the nodes.

    Instances of this class should be used as decorators for methods in subclasses of :class:`NodeAlgorithm`.
    """

    def __call__(self, func: ActionMethod):
        """
        Decorator that sets the method for the action and status.

        :param func: The method to be set.
        :type func: ActionMethod
        :return: The method that was set.
        :rtype: ActionMethod
        """

        assert (
            func.__name__ in Actions.__members__
        ), f"Invalid function name '{func.__name__}', please make sure it is one of {list(Actions.__members__)}."

        self.set_method(func)

        return func

    def implements(self, action: Actions):
        return hasattr(self, str(action))

    def get_method(self, action: Actions):
        return getattr(self, str(action))

    def set_method(self, method: ActionMethod):
        setattr(self, method.__name__, method)


class NodeAlgorithm(BaseAlgorithm):
    """
    Base class for distributed algorithms.

    In subclass, the following functions and attributes should be defined:

    - Class attribute Status - subclass of StatusValues that defines all the possible statuses of the nodes
      STATUS must be written at the bottom after all functions are defined
    - Initializer: (optionally) Pass INI message to nodes that should
      start the algorithm and set some attributes needed by the specific
      algorithm
    """

    INI = MessageMetaHeader.INITIALIZATION_MESSAGE

    class Status(StatusValues):
        "Example of StatusValues subclass that defines the possible statuses of the nodes."

        IDLE = "IDLE"

    algorithm_restrictions = (
        FiniteCommunicationDelays,
        LocalOrientation,
    )

    S_init = ()
    "Tuple of statuses that nodes should have at the beginning of the algorithm."

    S_term = ()
    "Tuple of statuses that nodes should have at the end of the algorithm."

    NODE_WRAPPER_MANAGER_TYPE: type[WrapperManager] = WrapperManager
    "Type of the node wrapper manager. Default is :class:`WrapperManager`."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._alarms: list[Alarm] = []
        "List of alarms set for the nodes."

        self.nwm = self.NODE_WRAPPER_MANAGER_TYPE(self.network)
        "Node wrapper manager."

    ### BaseAlgorithm interface methods ###

    def step(self, check_restrictions: bool, step: int):
        if not self.is_initialized():
            self.notify_observers(ObservableEvents.algorithm_started, self)
            self.initializer()
            if check_restrictions:
                self.check_restrictions()

            if not any(len(node.outbox) + len(node.inbox) > 0 for node in self.network.nodes()):
                logger.warning("Initializer didn't send any message.")
        else:
            self.network.communicate()
            self._process_alarms()
            for node in self.network.nodes_sorted():
                delay = self.network.behavioral_properties._get_node_processing_frequency(node)
                if step % delay == 0:
                    self._node_step(node)
                else:
                    logger.debug(f"Node {node._internal_id} skipped. Delay: {delay}. Current step: {step}.")

        self.notify_observers(ObservableEvents.step_done, self)

    def is_halted(self):
        """
        Check if the distributed algorithm has come to an end or deadlock,
        i.e. there are no messages to pass and no alarms set.

        A not-started algorithm is considered halted.

        :return: True if the algorithm is halted, False otherwise.
        :rtype: bool
        """
        return (
            all((len(node.inbox) == 0 and len(node.outbox) == 0) for node in self.network.nodes())
            and len(self._alarms) == 0
            and all(len(self.network.get_transit_messages(u, v)) == 0 for u, v in self.network.edges())
        )

    ### Algorithm methods ###

    def initializer(self):
        """
        Method used to initialize the algorithm. Is always called first.
        :class:`NodeAlgorithm` subclasses may want to reimplement it.

        Base implementation sends an INI message to the node with the lowest id and applies all restrictions.
        """
        logger.trace("Initializing algorithm {}.", self.name)
        self.apply_restrictions()

        node: "Node" = self.network.nodes_sorted()[0]
        node.push_to_inbox(Message(meta_header=NodeAlgorithm.INI))
        for node in self.network.nodes_sorted():
            node.status = self.Status.IDLE

    def send_msg(self, node_source: "NodeAccess", message: Message["_NodeWrapper"]):
        """
        Send a message to nodes listed in message's destination field.

        Note: Destination should be a list of nodes or one node.

        Update message's source field and inserts in node's outbox one copy
        of it for each destination.

        :param node_source: The node that sends the message.
        :type node_source: NodeAccess
        :param message: The message to be sent.
        :type message: Message
        """
        source_node: "Node" = node_source.unbox()  # unwrap the proxy

        message.source = source_node
        if not isinstance(message.destination, Iterable):
            message.destination = [message.destination]
        for destination_w in message.destination:
            destination: "Node" = destination_w.unbox()  # unwrap the proxy
            source_node.push_to_outbox(message.copy(), destination)

    def send(
        self,
        node_source: "NodeAccess",
        data: object,
        destination: Union["NeighborLabel", list["NeighborLabel"]],
        header: str = None,
    ):
        """
        Send content to destination, with header.

        :param node_w: Sender node
        :param data: Content to be received by destination nodes.
        :param destination: Destination node or list of destinations.
        :param header: Optional, header for the message.
        """
        self.send_msg(node_source, Message(data=data, header=header, destination=destination))

    ### Alarm methods ###

    def set_alarm(self, node_p: "NodeAccess", time: int, message: Message["_NodeWrapper"] | None = None):
        """
        Set an alarm for the node.
        One unit of time is one step of the algorithm.
        After the time has passed, the alarm will trigger and the message will be sent to the node.
        When will that message be processed is not guaranteed to be immediate.

        Returns the alarm that was set, useful for disabling it later.

        :param node_p: The node for which the alarm is set.
        :type node: NodeProxy
        :param time: The time after which the alarm will trigger.
        :type time: int
        :param message: The message to be sent when the alarm triggers.
        :type message: Message
        :return: The alarm that was set.
        :rtype: Alarm
        """
        assert time > 0, "Time must be greater than 0."

        if message is None:
            message = Message()
        message.meta_header = MessageMetaHeader.ALARM_MESSAGE
        message.destination = node_p.unbox()

        # Ensure that the node is a NodeAccess instance
        node_p = self.nwm.get_node_access(node_p)

        alarm = Alarm(time, message, node_p)
        self._alarms.append(alarm)
        return alarm

    def disable_all_node_alarms(self, node_p: "NodeAccess"):
        """
        Disable all alarms set for the node.

        :param node: The node for which the alarms are disabled.
        :type node: NodeProxy
        """
        to_delete_messages = [alarm.message for alarm in self._alarms if alarm.node == node_p]
        self._alarms = [alarm for alarm in self._alarms if alarm.node != node_p]

        for message in node_p.unbox().inbox.copy():
            if message in to_delete_messages and message in node_p.inbox:
                node_p.inbox.remove(message)

    def disable_alarm(self, alarm: Alarm):
        """
        Disable an alarm.

        :param alarm: The alarm to be disabled.
        :type alarm: Alarm
        """
        if alarm in self._alarms:
            self._alarms.remove(alarm)
        elif alarm.message in alarm.node.unbox().inbox:
            alarm.node.unbox().inbox.remove(alarm.message)

    def update_alarm_time(self, alarm: Alarm, time_diff: int):
        """
        Update the time of an alarm by adding a time difference.
        The time difference can be negative.

        :param alarm: The alarm for which the time is updated.
        :type alarm: Alarm
        :param time_diff: The time difference to be added to the alarm's time.
        :type time_diff: int
        """
        if alarm in self._alarms:
            alarm.time_left += time_diff
        elif alarm.message in alarm.node.unbox().inbox:
            alarm.node.unbox().inbox.remove(alarm.message)
            alarm.time_left += time_diff
            alarm.triggered = False
            self._alarms.append(alarm)
        else:
            raise ValueError("Alarm not found. Cannot update time. Please check if it has been triggered.")

    ### Methods for inbox blocking ###

    def block_inbox(self, node: "NodeAccess", filter: Callable[[Message["_NodeWrapper"]], bool]) -> object:
        """
        Block the inbox of a node, so that only messages that satisfy the filter can be processed.

        :param node: The node for which the inbox is blocked.
        :type node: NodeAccess
        :param filter: The filter that the messages must satisfy.
        :type filter: Callable[[Message], bool]
        """

        def adapted_filter(message: Message["Node"]):
            message = message.copy()

            if message.destination is not None:
                message.destination = node
            if message.source is not None:
                message.source = node._get_in_neighbor_proxy(message.source)

            return filter(message)

        return node.unbox().block_inbox(adapted_filter)

    def unblock_inbox(self, node: "NodeAccess", filter: object):
        """
        Unblock the inbox of a node, so that all messages can be processed.

        :param node: The node for which the inbox is unblocked.
        :type node: NodeAccess
        :param filter: The filter that was used to block the inbox.
        :type filter: object
        """
        node.unbox().unblock_inbox(filter)

    def close(self, node: "NodeAccess", neighbor: "NeighborLabel"):
        """
        Close the inbox of the node for a specific neighbor.

        Implementation detail
        ---------------------

        The inbox is closed by blocking all messages that have the neighbor as the source.
        This is done by adding a filter to the inbox that blocks all messages that have the neighbor as the source.
        To be able to unblock the inbox later, the filter is stored in the node's memory, under the key
        `__closed_edges_filters__`.

        :param node: The node for which the inbox is closed.
        :type node: NodeAccess
        :param neighbor: The neighbor for which the inbox is closed.
        :type neighbor: NeighborLabel
        """

        def filter_fn(message: Message["Node"]):
            return message.source != neighbor.unbox()

        if "__closed_edges_filters__" not in node.memory:
            node.memory["__closed_edges_filters__"] = defaultdict(list)

        node.memory["__closed_edges_filters__"][neighbor].append(filter_fn)
        node.unbox().block_inbox(filter_fn)

    def open(self, node: "NodeAccess", neighbor: "NeighborLabel"):
        """
        Open the inbox of the node for a specific neighbor.

        :param node: The node for which the inbox is opened.
        :type node: NodeAccess
        :param neighbor: The neighbor for which the inbox is opened.
        :type neighbor: NeighborLabel
        """
        if "__closed_edges_filters__" in node.memory:
            if neighbor in node.memory["__closed_edges_filters__"]:
                filters = node.memory["__closed_edges_filters__"][neighbor]
                for filter_fn in filters:
                    node.unbox().unblock_inbox(filter_fn)

                node.memory["__closed_edges_filters__"][neighbor] = []

    ### Methods for algorithm evaluation ###

    def check_algorithm_initialization(self):
        """
        Check if the algorithm's current state matches the expected initial state.

        This method depends on the correct configuration of the :attr:`S_init` class attribute of the algorithm.
        """
        for node in self.network.nodes():
            if node.status not in self.S_init:
                raise AlgorithmException(
                    f"Node status not initialized as expected. Found {node.status} instead of any of {self.S_init}."
                )

    def check_algorithm_termination(self):
        """
        Check if the algorithm's current state matches the expected termination state.

        This method depends on the correct configuration of the :attr:`S_term` class attribute of the algorithm.
        """
        for node in self.network.nodes():
            if node.status not in self.S_term:
                raise AlgorithmException(
                    f"Node status not terminated as expected. Found {node.status} instead of any of {self.S_term}."
                )

    ### Private methods for algorithm execution ###

    def _node_step(self, node: "Node"):
        """Executes one step of the algorithm for given node."""
        message: Message = node.receive()

        if message:
            logger.debug("Processing step at node {}.", node._internal_id)
            if message.destination is None or message.destination == node:
                # when destination is None it is broadcast message
                self._process_message(node, message)

    def _process_alarms(self):
        for alarm in self._alarms.copy():
            # copy to avoid errors produced by modifying the list while iterating
            alarm.time_left -= 1
            if alarm.time_left <= 0:
                self._alarms.remove(alarm)
                alarm.triggered = True
                alarm.node.unbox().push_to_inbox(alarm.message)

    def _process_message(self, node: "Node", message: Message["Node"]):
        logger.debug("Processing message: 0x%x" % id(message))
        method_name = MSG_META_HEADER_MAP[message.meta_header]
        return self._process_action(method_name, node, message)

    def _process_action(self, action: Actions, node: "Node", message: Message["Node"]):
        status: StatusValues = getattr(self.Status, node.status)

        node_view = self.nwm.get_node_access(node)  # TODO: set offuscation here
        if message.source is not None:
            message.source = node_view._get_in_neighbor_proxy(message.source)
        if message.destination is not None:
            message.destination = node_view

        message: Message["_NodeWrapper"]

        method_name = f"{action}_{status}"
        default_name = f"{Actions.default}_{status}"

        if hasattr(self, method_name):
            method = getattr(self, method_name)
            return method(node_view, message)
        elif hasattr(self, default_name):
            method = getattr(self, default_name)
            return method(node_view, message)

        logger.error(f"Method {self.name}.{action} not implemented for status {node.status}.")

    def reset(self):
        super().reset()
        self._alarms = []
        self.nwm = self.NODE_WRAPPER_MANAGER_TYPE(self.network)

    ### Metaclass methods ###

    def __configure_class__(clsname: str, bases: tuple[type], dct: dict):
        """
        Metaclass method that configures the class by adding methods for all possible actions and statuses.

        :param clsname: The name of the class.
        :type clsname: str
        :param bases: The base classes of the class.
        :type bases: tuple[type]
        :param dct: The dictionary of the class.
        :type dct: dict
        """

        # Search for the Status class in the bases or in the dct
        status_class = None
        if "Status" in dct:
            status_class: StatusValues = dct["Status"]
        else:
            for base in bases:
                if hasattr(base, "Status") and issubclass(base.Status, StatusValues):
                    status_class: StatusValues = base.Status
                    break
        assert status_class is not None, "Status class not found in bases or in dct."

        # Create methods for all possible actions and statuses
        for action in Actions:
            # Create a method for invocations like `self.default(node, message)`
            # CAUTION: the resolution of the method is deferred to the instance, super() will not work
            def __unresolved_action_method__(
                alg_instance: "NodeAlgorithm", node: "NodeAccess", message: Message["_NodeWrapper"]
            ):
                alg_instance._process_action(action, node, message)

            dct[action] = __unresolved_action_method__

            for status in status_class:
                if status.implements(action):
                    # Create a method for invocations like `self.default_IDLE(node, message)`
                    # As the method is created in the class, super() will work.
                    dct[f"{action}_{status}"] = status.get_method(action)



from numpy import array, concatenate, dot, sqrt

from pydistsim.algorithm.node_wrapper import SensorNodeAccess
from pydistsim.demo_algorithms.niculescu2003.floodingupdate import FloodingUpdate


class DVHop(FloodingUpdate):
    """
    Data is {landmark: [x,y,hop_count], ...}
    """

    required_params = ("truePositionKey", "hopsizeKey")
    NODE_ACCESS_TYPE = SensorNodeAccess

    def initiator_condition(self, node: "SensorNodeAccess"):
        node.memory[self.truePositionKey] = node.compositeSensor.read().get("TruePos", None)
        # true if node is one of the landmarks
        return node.memory[self.truePositionKey] is not None

    def initiator_data(self, node: "SensorNodeAccess"):
        return {node.id: concatenate((node.memory[self.truePositionKey][:2], [1]))}

    def handle_flood_message(self, node: "SensorNodeAccess", message):
        if self.dataKey not in node.memory:
            node.memory[self.dataKey] = {}
        updated_data = {}
        for landmark_id, landmark_data in list(message.data.items()):
            # skip if landmark in message data is current node
            if landmark_id == node.id:
                continue
            # update only if this is first received data from landmark or new
            # hopcount is smaller than previous minimum
            if (
                landmark_id not in node.memory[self.dataKey]
                or landmark_data[2] < node.memory[self.dataKey][landmark_id][2]
            ):
                node.memory[self.dataKey][landmark_id] = array(landmark_data)
                # increase hopcount
                landmark_data[2] += 1
                updated_data[landmark_id] = landmark_data

        # if node is one of the landmarks then it should recalculate hopsize
        if node.memory[self.truePositionKey] is not None:
            self.recalculate_hopsize(node)

        return updated_data

    def recalculate_hopsize(self, node: SensorNodeAccess):
        pos = node.memory[self.truePositionKey]
        try:
            landmarks_count = len(node.memory[self.dataKey])
        except KeyError:
            pass
        else:

            def dist(x, y):
                return sqrt(dot(x - y, x - y))

            if landmarks_count > 0:
                node.memory[self.hopsizeKey] = sum(
                    [dist(lp[:2], pos) for lp in list(node.memory[self.dataKey].values())]
                ) / sum([lp[2] for lp in list(node.memory[self.dataKey].values())])



from pydistsim.algorithm import NodeAlgorithm, StatusValues
from pydistsim.message import Message
from pydistsim.restrictions.communication import BidirectionalLinks
from pydistsim.restrictions.reliability import TotalReliability
from pydistsim.restrictions.topological import Connectivity


class FloodingUpdate(NodeAlgorithm):
    """
    This is modified Flooding algorithm (Santoro2007 p.13) so that every node
    continues to forward flood messages as long as information gathered is
    updating its knowledge.
    Note: does not have global termination detection
    Costs: ?
    """

    required_params = ("dataKey",)  # memory key for data being updated
    default_params = {}

    class Status(StatusValues):
        FLOODING = "FLOODING"
        INITIATOR = "INITIATOR"

    S_init = (Status.INITIATOR, Status.FLOODING)
    S_term = (Status.FLOODING,)

    algorithm_restrictions = (
        BidirectionalLinks,
        TotalReliability,
        Connectivity,
    )

    def initializer(self):
        """
        Starts in every node satisfying initiator condition.
        """

        for node in self.network.nodes():
            if self.initiator_condition(node):
                node.push_to_inbox(Message(destination=node, meta_header=NodeAlgorithm.INI))
                node.status = self.Status.INITIATOR
            node.status = self.Status.FLOODING

    @Status.INITIATOR
    def spontaneously(self, node, message):
        self.send_msg(node, Message(header="Flood", data=self.initiator_data(node)))

    @Status.FLOODING
    def receiving(self, node, message):
        updated_data = self.handle_flood_message(node, message)
        if updated_data:
            self.send_msg(node, Message(header="Flood", data=updated_data))

    def initiator_condition(self, node):
        raise NotImplementedError

    def initiator_data(self, node):
        raise NotImplementedError

    def handle_flood_message(self, node, message):
        raise NotImplementedError



import unittest

import scipy.stats

from pydistsim.network import Network, Node
from pydistsim.network.sensor import DistSensor, SensorError


class TestSensor(unittest.TestCase):

    def test_read(self):
        """Test read compositeSensor"""
        net = Network()
        node = net.add_node()
        node.compositeSensor.read()

    def test_set_compositeSensor(self):
        """Test setting compositeSensors on a node"""
        node = Node()
        dist_sensor = DistSensor({"pf": scipy.stats.norm, "scale": 10})
        node.compositeSensor = ("AoASensor", dist_sensor)
        self.assertRaises(SensorError, node.compositeSensor.read)

        net = Network()
        node = net.add_node()
        dist_sensor = DistSensor({"pf": scipy.stats.norm, "scale": 10})
        node.compositeSensor = ("AoASensor", dist_sensor)
        self.assertTrue(len(node.compositeSensor.sensors) == 2)
        readings = node.compositeSensor.read()
        self.assertTrue("AoA" in list(readings.keys()) and "Dist" in list(readings.keys()))

        # TODO: check normal distribution



"""
Algorithms behavior and base classes.
"""

# flake8: noqa: F401

from pydistsim.algorithm.base_algorithm import AlgorithmException, BaseAlgorithm
from pydistsim.algorithm.network_algorithm import NetworkAlgorithm
from pydistsim.algorithm.node_algorithm import Actions, NodeAlgorithm, StatusValues



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional

from numpy import allclose, sign, sqrt
from numpy.random import random

from pydistsim.logging import logger
from pydistsim.network.environment import Environment
from pydistsim.network.network import BidirectionalNetwork, Network
from pydistsim.utils.helpers import with_typehint

if TYPE_CHECKING:
    from pydistsim.network.behavior import NetworkBehaviorModel
    from pydistsim.network.node import Node


class RangeType(ABC):
    """RangeType abstract base class.

    This class represents an abstract base class for different types of channels.
    Subclasses of RangeType should implement the `in_comm_range` method.

    :param environment: The environment in which the channel operates.
    :type environment: Environment
    """

    def __init__(self, environment: "Environment"):
        self.environment = environment

    @abstractmethod
    def in_comm_range(self, network: "RangeNetworkType", node1: "Node", node2: "Node"):
        """Check if two nodes are within communication range.

        This method should be implemented by subclasses to determine if two nodes
        are within communication range.

        :param network: The network in which the nodes are connected.
        :type network: Network
        :param node1: The first node.
        :type node1: Node
        :param node2: The second node.
        :type node2: Node
        :return: True if the nodes are within communication range, False otherwise.
        :rtype: bool
        """
        ...


class UdgRangeType(RangeType):
    """Unit disc graph range type.

    This class represents the Unit Disc Graph (UDG) channel type. It determines if
    two nodes are within communication range based on their positions and communication
    range.

    :param environment: The environment in which the channel operates.
    :type environment: Environment
    """

    def in_comm_range(self, network: "RangeNetworkType", node1: "Node", node2: "Node"):
        """Check if two nodes are within communication range.

        Two nodes are in communication range if they can see each other and are
        positioned so that their distance is smaller than the communication range.

        :param network: The network in which the nodes are connected.
        :type network: Network
        :param node1: The first node.
        :type node1: Node
        :param node2: The second node.
        :type node2: Node
        :return: True if the nodes are within communication range, False otherwise.
        :rtype: bool
        """
        p1 = network.pos[node1]
        p2 = network.pos[node2]
        d = sqrt(sum(pow(p1 - p2, 2)))
        if d < node1.commRange and d < node2.commRange:
            if self.environment.are_visible(p1, p2):
                return True
        return False


class CompleteRangeType(RangeType):
    """Complete range type.

    This class represents the Complete channel type. It always returns True,
    indicating that any two nodes are within communication range.

    :param environment: The environment in which the channel operates.
    :type environment: Environment
    """

    def in_comm_range(self, network: "RangeNetworkType", node1: "Node", node2: "Node"):
        """Check if two nodes are within communication range.

        This method always returns True, indicating that any two nodes are within
        communication range.

        :param network: The network in which the nodes are connected.
        :type network: Network
        :param node1: The first node.
        :type node1: Node
        :param node2: The second node.
        :type node2: Node
        :return: True if the nodes are within communication range, False otherwise.
        :rtype: bool
        """
        return True


class SquareDiscRangeType(RangeType):
    """Square Disc channel type.

    This class represents the Square Disc channel type. It determines if two nodes
    are within communication range based on their positions, communication range,
    and a probability of connection.

    :param environment: The environment in which the channel operates.
    :type environment: Environment
    """

    def in_comm_range(self, network: "RangeNetworkType", node1: "Node", node2: "Node"):
        """Check if two nodes are within communication range.

        Two nodes are in communication range if they can see each other, are positioned
        so that their distance is smaller than the communication range, and satisfy a
        probability of connection.

        :param network: The network in which the nodes are connected.
        :type network: Network
        :param node1: The first node.
        :type node1: Node
        :param node2: The second node.
        :type node2: Node
        :return: True if the nodes are within communication range, False otherwise.
        :rtype: bool
        """
        p1 = network.pos[node1]
        p2 = network.pos[node2]
        d = sqrt(sum(pow(p1 - p2, 2)))
        if random() > d**2 / node1.commRange**2:
            if self.environment.are_visible(p1, p2):
                assert node1.commRange == node2.commRange
                return True
        return False


class RangeNetworkMixin(with_typehint(Network)):
    """
    Mixin to define a type of network that decides which nodes are connected based on their
    communication range.


    :param environment: The environment in which the network operates. If not provided, a new Environment instance will be created.
    :type environment: Environment, optional
    :param rangeType: The type of channel to be used for communication. If not provided, a new RangeType instance will be created using the environment.
    :type rangeType: RangeType, optional
    :param graph: The graph representing the network topology. Defaults to None.
    :type graph: NetworkX graph, optional
    :param kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        incoming_graph_data=None,
        environment: Optional["Environment"] = None,
        rangeType: RangeType | None = None,
        behavioral_properties: Optional["NetworkBehaviorModel"] = None,
        **kwargs,
    ):
        super().__init__(incoming_graph_data, environment, behavioral_properties, **kwargs)
        self.rangeType = rangeType or UdgRangeType(self._environment)
        self.rangeType.environment = self._environment

    @staticmethod
    def to_directed_class():
        return RangeNetwork

    @staticmethod
    def to_undirected_class():
        return BidirectionalRangeNetwork

    def __deepcopy__(self, memo, nodes=None, edges=None, init_args=None, cls=None):
        init_args = init_args or {}
        init_args["rangeType"] = self.rangeType

        return super().__deepcopy__(memo, nodes, edges, init_args, cls)

    def add_node(self, node=None, pos=None, ori=None, commRange=None):
        node = super().add_node(node, pos, ori, commRange)
        self.recalculate_edges([node])
        return node

    def remove_node(self, node, skip_check=False):
        super().remove_node(node, skip_check)
        self.recalculate_edges()

    def recalculate_edges(self, nodes: Iterable | None = None):
        """
        Recalculate edges for given nodes or for all self.nodes().

        :param nodes: A list of nodes to recalculate edges for. If not provided, edges will be recalculated for all nodes in the network.
        :type nodes: list, optional

        Edge between nodes n1 and n2 are added if both are RangeType.in_comm_range of each other.
        """
        if not nodes:
            nodes = self.nodes()
        for n1 in nodes:
            for n2 in self.nodes():
                if n1 != n2:
                    for x, y in ((n1, n2), (n2, n1)):
                        if self.rangeType.in_comm_range(self, x, y):
                            super().add_edge(x, y)
                        elif self.has_edge(x, y):
                            self.remove_edge(x, y)

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        """
        Add an edge to the network.

        :param u_of_edge: The source node of the edge.
        :param v_of_edge: The target node of the edge.
        :param attr: Additional attributes to be assigned to the edge.
        """
        logger.warning("Edges are auto-calculated from rangeType and commRange")
        super().add_edge(u_of_edge, v_of_edge, **attr)

    def _set_environment(self, environment: Environment):
        super()._set_environment(environment)
        self.rangeType.environment = environment
        for node in self.nodes_sorted():
            self.remove_node(node, skip_check=True)
            self.add_node(node)
        logger.debug("All nodes are moved into new environment.")

    def validate_params(self, params: dict):
        super().validate_params(params)
        for param, value in params.items():
            if param == "rangeType":
                assert self.rangeType.__class__ == value.__class__
            elif param == "comm_range":
                for node in self:
                    assert node.commRange == value

    def modify_avg_degree(self, value):
        """
        DEPRECATED AND UNTESTED

        Modifies (increases) average degree based on the given value by
        modifying nodes' commRange.

        :param value: The desired average degree value.
        :type value: float

        :raises AssertionError: If all nodes do not have the same commRange.
        :raises AssertionError: If the given value is not greater than the current average degree.

        This method increases the average degree of the network by modifying the communication range
        (`commRange`) of the nodes. It ensures that all nodes have the same communication range.

        The method uses a step-wise approach to gradually increase the average degree until it reaches
        the desired value. It adjusts the communication range of each node in the network by adding a
        step size calculated based on the difference between the desired average degree and the current
        average degree.

        The step size is determined by the `step_factor` parameter, which controls the rate of change
        in the communication range. If the step size overshoots or undershoots the desired average
        degree, the `step_factor` is halved to reduce the step size for the next iteration.

        Once the average degree reaches the desired value, the method logs the modified degree.

        Note: This method assumes that the network is initially connected and all nodes have the same
        communication range.

        Example usage:
            network.modify_avg_degree(5.0)
        """
        # assert all nodes have the same commRange
        assert allclose([n.commRange for n in self], self.nodes_sorted()[0].commRange)
        # TODO: implement decreasing of degree, preserve connected network
        assert value + 1 > self.avg_degree()  # only increment
        step_factor = 7.0
        steps = [0]
        # TODO: while condition should call validate
        while not allclose(self.avg_degree(), value, atol=1):
            steps.append((value - self.avg_degree()) * step_factor)
            for node in self:
                node.commRange += steps[-1]
            # variable step_factor for step size for over/undershoot cases
            if len(steps) > 2 and sign(steps[-2]) != sign(steps[-1]):
                step_factor /= 2
        logger.trace("Modified degree to {}", self.avg_degree())


class RangeNetwork(RangeNetworkMixin, Network):
    """
    Type of network that decides which nodes are connected based on their communication range.

    Aims to represent a wireless network where nodes can only communicate with each other if they
    are within a certain range.

    Manual edge modification is not recommended. Edges are automatically calculated and any edge
    can be removed by moving the nodes out of communication range or by addition/removal of nodes.
    """


class BidirectionalRangeNetwork(RangeNetworkMixin, BidirectionalNetwork):
    """
    Same as RangeNetwork but with bidirectional edges (undirected graph).
    """


RangeNetworkType = RangeNetwork | BidirectionalRangeNetwork
"Type of network that decides which nodes are connected based on their communication range."



def start_ipydistsim():
    import os
    import sys

    if sys.platform.startswith("win32") or sys.platform.startswith("linux2"):
        try:
            os.environ["VIRTUAL_ENV"] = os.environ["PYDISTSIM_ENV"]
            os.environ["IPYTHONDIR"] = os.path.join(os.environ["PYDISTSIM_ENV"], ".ipython")
        except KeyError:
            pass

        from IPython.frontend.terminal.ipapp import TerminalIPythonApp

        app = TerminalIPythonApp.instance()
        app.initialize(argv=["--profile=pydistsim"])
        app.start()

    # TODO: support other platforms
    # elif sys.platform.startswith("darwin"):



import os
import sys

from setuptools import setup

# transfer profile_pydistsim for ipython into IPYTHONDIR
if "install" in sys.argv or "develop" in sys.argv:
    import shutil

    try:
        from IPython.paths import get_ipython_dir

        ipythondir = get_ipython_dir()
    except ImportError:  # @ReservedAssignment
        print(
            "PyDistSim IPython configuration not installed. Install latest "
            "IPython and then copy the conf/ipython/profile_pydistsim/"
            "ipython_config.py manually to IPython config dir."
        )
    else:
        profiledir = os.path.join(ipythondir, "profile_pydistsim")
        if not os.path.exists(ipythondir):
            os.makedirs(ipythondir)
        if not os.path.exists(profiledir):
            os.makedirs(profiledir)
        print("copying ipython_config.py and ipython_notebook_config.py to " + profiledir)
        shutil.copy(
            os.path.join("pydistsim", "conf", "ipython", "ipython_config.py"),
            profiledir,
        )
        shutil.copy(
            os.path.join("pydistsim", "conf", "ipython", "ipython_notebook_config.py"),
            profiledir,
        )

setup()



"""
This module provides model for QTreeView widget that is created out of
dictionary data.
"""

from PySide6.QtCore import QAbstractItemModel, QModelIndex, Qt


class DictionaryTreeModel(QAbstractItemModel):
    def __init__(self, parent=None, dic={}):
        super().__init__(parent)
        self.dic = dic
        self.rootItem = TreeItem(("tree", "root"), None)
        self.parents = {0: self.rootItem}
        self.setupModelData()

    def setupModelData(self):
        items = list(self.dic.items())
        items.sort()
        for item in items:
            newparent = TreeItem(item, self.rootItem)
            self.rootItem.appendChild(newparent)

    def columnCount(self, parent=None):
        if parent and parent.isValid():
            return parent.internalPointer().columnCount()
        else:
            return 1

    def data(self, index, role):
        if not index.isValid():
            return None

        item = index.internalPointer()
        if role == Qt.DisplayRole:
            return item.data(index.column())
        if role == Qt.UserRole:
            if item:
                return item.data(0)

        return None

    def index(self, row, column, parent):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        if not childItem:
            return QModelIndex()

        parentItem = childItem.parent()

        if parentItem == self.rootItem:
            return QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent=QModelIndex()):
        if parent.column() > 0:
            return 0
        if not parent.isValid():
            p_Item = self.rootItem
        else:
            p_Item = parent.internalPointer()
        return p_Item.childCount()


class TreeItem:
    """
    a python object used to return row/column data, and keep note of
    it's parents and/or children
    """

    def __init__(self, dicItem, parentItem):
        key, value = dicItem
        self.itemDataKey = key
        self.itemDataValue = value
        self.parentItem = parentItem
        self.childItems = []
        if isinstance(value, dict):
            self.itemData = key
            items = list(value.items())
            items.sort()
            for item in items:
                self.appendChild(TreeItem(item, self))
        # TODO: solve infinite recursion problem
        # elif hasattr(value,'get_dic'):
        #    self.itemData = key
        #    items = value.get_dic().items()
        #    items.sort()
        #    for item in items:
        #        self.appendChild(TreeItem(item,self))
        else:
            self.itemData = ": ".join([key.__str__(), value.__str__()])

    def appendChild(self, item):
        self.childItems.append(item)

    def child(self, row):
        return self.childItems[row]

    def childCount(self):
        return len(self.childItems)

    def columnCount(self):
        return 1

    def data(self, column):
        return self.itemData

    def parent(self):
        return self.parentItem

    def row(self):
        if self.parentItem:
            return self.parentItem.childItems.index(self)
        return 0

    def toString(self, indentation=""):
        """
        Return string that represents all data in this item and all children.
        """
        s = self.itemData
        for child in self.childItems:
            s += "\n" + indentation + child.toString(indentation + "    ")
        return s



__all__ = ["read_pickle", "write_pickle"]

import errno
import os
import pickle as pickle
from types import MethodType
from typing import Any

from pydistsim.logging import logger


def _get_fh(path: str, mode="r"):
    """Return a file handle for given path and attempt to decompress/compress
    files ending in '.gz'

    :param path: The path of the file.
    :type path: str
    :param mode: The mode in which the file should be opened. Default is 'r'.
    :type mode: str
    :return: The file handle.
    :rtype: file object
    """

    if path.endswith(".gz"):
        import gzip

        fh = gzip.open(path, mode=mode)
    else:
        fh = open(path, mode=mode)
    return fh


def write_pickle(obj: Any, path: str, makedir=True):
    """Write object in Python pickle format.

    :param obj: The object to be pickled.
    :type obj: Any
    :param path: The path of the file to save the pickled object.
    :type path: str
    :param makedir: Whether to create the directory if it doesn't exist. Default is True.
    :type makedir: bool
    """

    try:
        os.makedirs(os.path.split(path)[0])
    except OSError as e:
        if e.errno != errno.EEXIST and e.filename != "":
            raise e
    try:
        fh = _get_fh(str(path), mode="wb")
        pickle.dump(obj, fh, pickle.HIGHEST_PROTOCOL)
    finally:
        fh.close()
    logger.info(f"Instance of {str(obj.__class__)} saved in {path}")


write_npickle = write_pickle


def read_pickle(path: str, not_found_raises=True):
    """Read object in Python pickle format. If not_found_raises is True then raise
    an exception if file is missing.

    :param path: The path of the file to read the pickled object from.
    :type path: str
    :param not_found_raises: Whether to raise an exception if the file is missing. Default is True.
    :type not_found_raises: bool
    :return: The unpickled object.
    :rtype: Any
    """

    try:
        fh = _get_fh(str(path), "rb")
        obj = pickle.load(fh)
        logger.info(f"Instance of {str(obj.__class__)} loaded: {path}")
        return obj
    except OSError as e:
        # if error is some other than errno.ENOENT ='file not found raise
        if not_found_raises or e.errno != errno.ENOENT:
            raise e
        return None


read_npickle = read_pickle


# scipy.stats.norm (scipy.stats.distributions.norm_gen) object has some bounded
# (instance) methods that needs to be pickled
# this is solution for pickling instance methods found at
# http://stackoverflow.com/a/1816969/1247955
def _pickle_method(method: MethodType):
    """Pickle a bound method.

    :param method: The method to be pickled.
    :type method: MethodType
    :return: The pickled method.
    :rtype: tuple
    """

    func_name = method.__func__.__name__
    obj = method.__self__
    cls = method.__self__.__class__
    return _unpickle_method, (func_name, obj, cls)


def _unpickle_method(func_name: str, obj: Any, cls: type):
    """Unpickle a bound method.

    :param func_name: The name of the method.
    :type func_name: str
    :param obj: The object to which the method belongs.
    :type obj: Any
    :param cls: The class of the object.
    :type cls: type
    :return: The unpickled method.
    :rtype: method
    """

    for cls in cls.mro():
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)


import copyreg

copyreg.pickle(MethodType, _pickle_method, _unpickle_method)



"""
Behavioral properties for a network. These properties include message ordering, message loss, clock increment, message
delay, node processing frequency and bounded communication delays. The properties are used to simulate different
network behaviors, such as a network with message ordering, no message loss and synchronized clocks.
"""

from collections.abc import Callable
from dataclasses import dataclass
from random import randint, uniform
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from pydistsim.message import Message
    from pydistsim.network.network import NetworkType
    from pydistsim.network.node import Node


#### Delay functions ####


def delay_size_network(network: "NetworkType", message: "Message") -> int:
    "Delay every message by the number of nodes in the network."
    return len(network)


def random_delay_max_size_network(network: "NetworkType", message: "Message") -> int:
    "Delay a message by a random number between 0 and the number of nodes in the network."
    return randint(0, len(network))


def delay_based_on_network_usage(network: "NetworkType", message: "Message") -> int:
    "Delay a message by the number of pending messages in the network."
    # 30 messages in a 10 node network == 3 steps delay
    return round(sum(len(node.outbox) for node in network) / len(network))


#### Loss functions ####


def random_loss(probability_of_loss) -> Callable[["NetworkType", "Message"], bool]:
    "Randomly lose a message with a given probability."

    def _random_loss(network: "NetworkType", message: "Message") -> bool:
        return uniform(0, 1) < probability_of_loss

    return _random_loss


#### Clock increments ####


def big_random_increment(node: "Node") -> int:
    "Random increment between 1 and 10."
    return randint(1, 10)


def small_random_increment(node: "Node") -> int:
    "Random increment is 1 or 2."
    return randint(1, 2)


#### Node delay ####


def half_the_nodes_are_slow(node: "Node") -> int:
    "Half the nodes are slow and will process every other message."
    return 2 if node._internal_id % 2 == 0 else 1


def half_the_nodes_are_very_slow(node: "Node") -> int:
    "Half the nodes are fast and will process every message."
    return 4 if node._internal_id % 2 == 0 else 1


@dataclass(slots=True, frozen=True)
class NetworkBehaviorModel:
    """
    Behavioral properties for a network.

    :param message_ordering: Boolean indicating if messages should be ordered. Default is False.
    :param message_loss_indicator: Function that returns a boolean indicating if a given message should be lost.
        Default (None) means no message loss.
    :param clock_increment: Function that returns the increment for the clock of a given node. Default (None) means
        unitary clock increment (sync).
    :param node_processing_frequency: Function that controls how often a node may process a message. Default (None)
        means no delay. The integer must be greater than 0. For example, if the integer is 2, the node will process every
        other message. As this will be called on every simulation step, it is recommended to use a constant value or a
        function that does not depend on the node state, since doing so may lead to unexpected behavior, such as a node
        never processing any message.
    :param message_delay_indicator: Function that returns the amount of steps to delay a given message. Default (None)
        means no message delay.
    :param bounded_communication_delays: Boolean indicating if there is a predefined constant T such that the
        communication delay of any message on any link is at most T. Default is False.
    """

    message_ordering: bool = False
    "Boolean indicating if messages should be ordered."

    message_loss_indicator: Callable[["NetworkType", "Message"], bool] | None = None
    "Function that returns a boolean indicating if a given message should be lost. None means no message loss."

    clock_increment: Callable[["Node"], int] | None = None
    "Function that returns the increment for the clock of a given node. None means unitary clock increment (sync)."

    node_processing_frequency: Callable[["Node"], int] | None = None
    """
    Function that controls how often a node may process a message. None means no delay. The integer must be greater
    than 0. For example, if the integer is 2, the node will process every other message. As this will be called on every
    simulation step, it is recommended to use a constant value or a function that does not depend on the node state,
    since doing so may lead to unexpected behavior, such as a node never processing a message.
    """

    message_delay_indicator: Callable[["NetworkType", "Message"], int] | None = None
    "Function that returns the amount of steps to delay a given message. None means no message delay."

    bounded_communication_delays: bool = False
    "If there is a predefined constant T such that the communication delay of any message on any link is at most T."

    ######## Methods ########

    def _get_delay(self, network: "NetworkType", message: "Message") -> int:
        "Get the delay for a given message."
        return self.message_delay_indicator(network, message) if self.message_delay_indicator else 0

    def _should_lose(self, network: "NetworkType", message: "Message") -> bool:
        "Check if a given message should be lost."
        return self.message_loss_indicator(network, message) if self.message_loss_indicator else False

    def _get_clock_increment(self, node: "Node") -> int:
        "Get the increment for the clock of a given node."
        return self.clock_increment(node) if self.clock_increment else 1

    def _get_node_processing_frequency(self, node: "Node") -> int:
        "Get the delay for a given node."
        if self.node_processing_frequency is None:
            return 1

        delay = self.node_processing_frequency(node)
        assert delay > 0, "Node processing delay must be greater than 0."
        return delay

    #### Communication properties example instances ####

    IdealCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, no message loss, no message delay and synchronized clocks."

    UnorderedCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with no message ordering, no message loss, no message delay and unsynchronized clocks."

    ThrottledCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, no message loss, a delay based on network usage and unsynchronized clocks."

    UnorderedThrottledCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with no message ordering, no message loss, a delay based on network usage and unsynchronized clocks."

    RandomDelayCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, no message loss, random delay based on the network size and unsynchronized clocks."

    RandomDelayCommunicationSlowNodes: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, no message loss, random delay based on the network size, unsynchronized clocks and half the nodes are 'slow'."

    RandomDelayCommunicationVerySlowNodes: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, no message loss, random delay based on the network size, unsynchronized clocks and half the nodes are very 'slow'."

    UnorderedRandomDelayCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with no message ordering, no message loss, random delay based on the network size and unsynchronized clocks."

    UnlikelyRandomLossCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, a random (but unlikely) message loss, no message delay and unsynchronized clocks."

    LikelyRandomLossCommunication: ClassVar["NetworkBehaviorModel"]
    "Properties for a network with message ordering, a random (but likely) message loss, no message delay and unsynchronized clocks."


NetworkBehaviorModel.IdealCommunication = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=None,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=None,
)

NetworkBehaviorModel.UnorderedCommunication = NetworkBehaviorModel(
    message_ordering=False,
    message_delay_indicator=None,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
)

NetworkBehaviorModel.ThrottledCommunication = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=delay_based_on_network_usage,
    bounded_communication_delays=False,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
)

NetworkBehaviorModel.UnorderedThrottledCommunication = NetworkBehaviorModel(
    message_ordering=False,
    message_delay_indicator=delay_based_on_network_usage,
    bounded_communication_delays=False,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
)

NetworkBehaviorModel.RandomDelayCommunication = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=random_delay_max_size_network,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
)

NetworkBehaviorModel.RandomDelayCommunicationSlowNodes = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=random_delay_max_size_network,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
    node_processing_frequency=half_the_nodes_are_slow,
)

NetworkBehaviorModel.RandomDelayCommunicationVerySlowNodes = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=random_delay_max_size_network,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
    node_processing_frequency=half_the_nodes_are_very_slow,
)

NetworkBehaviorModel.UnorderedRandomDelayCommunication = NetworkBehaviorModel(
    message_ordering=False,
    message_delay_indicator=random_delay_max_size_network,
    bounded_communication_delays=True,
    message_loss_indicator=None,
    clock_increment=small_random_increment,
)

NetworkBehaviorModel.UnlikelyRandomLossCommunication = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=None,
    bounded_communication_delays=True,
    message_loss_indicator=random_loss(0.1),
    clock_increment=small_random_increment,
)

NetworkBehaviorModel.LikelyRandomLossCommunication = NetworkBehaviorModel(
    message_ordering=True,
    message_delay_indicator=None,
    bounded_communication_delays=True,
    message_loss_indicator=random_loss(0.9),
    clock_increment=small_random_increment,
)



import unittest
from copy import deepcopy

from pydistsim._exceptions import SimulationException
from pydistsim.algorithm import NetworkAlgorithm, NodeAlgorithm
from pydistsim.network import NetworkGenerator
from pydistsim.simulation import Simulation
from pydistsim.utils.testing import PyDistSimTestCase


class UnimplementedNodeAlgorithm(NodeAlgorithm): ...


class ImplementedNetworkAlgorithm(NetworkAlgorithm):

    def run(self):
        for node in self.network.nodes():
            node.memory["test"] = "test"


class TestRunBaseNodeAlgorithm(PyDistSimTestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100)
        self.net = net_gen.generate_random_network()
        self.algorithms = (UnimplementedNodeAlgorithm,)

    def test_run_base_algorithm(self):
        sim = Simulation(self.net)
        sim.algorithms = self.algorithms

        for node in self.net.nodes():
            with self.subTest(node=node):
                print(f"{node._internal_id=}, {node.status=}, {node.outbox=}, {node.inbox=}")
                assert len(node.outbox) == 0
                assert len(node.inbox) == 0

        assert sim.is_halted()

        sim.run_step()
        # First step should put the INI in the outbox
        assert all(len(node.outbox) == 0 for node in self.net.nodes())
        assert sum(len(node.inbox) for node in self.net.nodes()) == 1

        # Put the INI message in the inbox of a node
        sim.run_step()

        nodes_with_1_msg = 0
        for node in self.net.nodes():
            with self.subTest(node=node):
                print(f"{node._internal_id=}, {node.status=}, {node.outbox=}, {node.inbox=}")
                assert len(node.outbox) == 0
                nodes_with_1_msg += 1 if len(node.inbox) else 0
        assert nodes_with_1_msg == 1

        assert not sim.is_halted()

        sim.run_step()
        # Second step should process the INI message (and do nothing)

        assert all([len(node.outbox) == 0 for node in self.net.nodes()])
        assert all([len(node.inbox) == 0 for node in self.net.nodes()])

        assert sim.is_halted()

    def test_get_current_algorithm(self):
        """Test getting current algorithm."""
        sim = Simulation(self.net)
        sim.algorithms = ()

        with self.assertRaises(SimulationException):
            sim.get_current_algorithm()

        sim.algorithms = (NodeAlgorithm,)


class TestRunNetworkAlgorithm(unittest.TestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100)
        self.net = net_gen.generate_random_network()
        self.algorithms = (ImplementedNetworkAlgorithm,)

    def test_run_base_algorithm(self):
        sim = Simulation(self.net)
        sim.algorithms = self.algorithms

        for node in self.net.nodes():
            with self.subTest(node=node):
                print(f"{node._internal_id=}, {node.status=}, {node.outbox=}, {node.inbox=}")
                assert "test" not in node.memory

        sim.run(1)

        for node in self.net.nodes():
            with self.subTest(node=node):
                print(f"{node._internal_id=}, {node.status=}, {node.outbox=}, {node.inbox=}")
                assert node.memory["test"] == "test"

        assert sim.is_halted()


class TestRunNotImplementedNetworkAlgorithm(unittest.TestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100)
        self.net = net_gen.generate_random_network()
        self.algorithms = (NetworkAlgorithm,)

    def test_run_base_algorithm(self):
        sim = Simulation(self.net)
        sim.algorithms = self.algorithms

        with self.assertRaises(NotImplementedError):
            sim.run(100_000)

        assert sim.is_halted()


class TestResetNetwork(unittest.TestCase):

    def setUp(self):
        net_gen = NetworkGenerator(100)
        self.net1 = net_gen.generate_random_network()
        self.sim1 = Simulation(self.net1)
        self.sim1.algorithms = (UnimplementedNodeAlgorithm,)

        self.net2 = net_gen.generate_random_network()
        self.sim2 = Simulation(self.net2)
        self.sim2.algorithms = (ImplementedNetworkAlgorithm,)

    def test_set_network(self):
        assert isinstance(self.sim1.get_current_algorithm(), UnimplementedNodeAlgorithm)
        assert isinstance(self.sim2.get_current_algorithm(), ImplementedNetworkAlgorithm)

        assert self.sim1.network == self.net1
        assert self.net1.simulation == self.sim1

        self.sim1.network = self.net2

        assert self.sim1.network == self.net2
        assert self.net2.simulation == self.sim1
        assert self.net1.simulation is None

    def test_deepcopy(self):
        sim_copy = deepcopy(self.sim1)

        assert sim_copy is not self.sim1
        assert sim_copy.network is not self.sim1.network
        assert len(sim_copy.network) == len(self.net1)
        assert len(sim_copy.network.edges()) == len(self.net1.edges())
        assert isinstance(sim_copy.network, type(self.net1))
        assert len(sim_copy.algorithms) == len(self.sim1.algorithms)
        assert all(
            isinstance(copy_algo, type(og_algo))
            for copy_algo, og_algo in zip(sim_copy.algorithms, self.sim1.algorithms)
        )



"""
Complete source for the pydistsim package.
"""

# flake8: noqa: F401

from importlib import metadata

try:
    package_metadata = metadata.metadata("pydistsim")

    __author__ = package_metadata["author-email"]
    __version__ = package_metadata["version"]
except metadata.PackageNotFoundError:
    __author__ = __version__ = None

# For interactive sessions these import names with from pydistsim import *
import os

try:
    from PySide6.QtCore import SIGNAL as __SIGNAL

    os.environ["QT_API"] = "pyside"
except ImportError:
    "No PySide6 found."
    ...

# Declare namespace package
from pkgutil import extend_path  # @Reimport

from pydistsim.algorithm import NodeAlgorithm, StatusValues
from pydistsim.benchmark import AlgorithmBenchmark
from pydistsim.logging import disable_logger, enable_logger, logger
from pydistsim.network import BidirectionalNetwork, Network, NetworkGenerator, Node
from pydistsim.simulation import Simulation

__path__ = extend_path(__path__, __name__)  # @ReservedAssignment



from collections.abc import Iterable
from copy import deepcopy
from math import inf as Inf
from random import choices
from typing import TYPE_CHECKING

from networkx import (
    DiGraph,
    Graph,
    draw_networkx_edges,
    draw_networkx_labels,
    draw_networkx_nodes,
    is_connected,
    is_strongly_connected,
)
from numpy import allclose, array, average, max, min, pi
from numpy.random import rand

from pydistsim._exceptions import MessageUndeliverableException, NetworkException
from pydistsim.gui import drawing as draw
from pydistsim.logging import logger
from pydistsim.network.behavior import NetworkBehaviorModel
from pydistsim.network.environment import Environment, Environment2D
from pydistsim.network.node import Node
from pydistsim.network.sensor import CompositeSensor
from pydistsim.observers import NodeObserver, ObserverManagerMixin
from pydistsim.utils.helpers import pydistsim_equal_objects, with_typehint

if TYPE_CHECKING:
    from pydistsim.message import Message


class NetworkMixin(ObserverManagerMixin, with_typehint(Graph)):
    """
    Mixin to extend Graph and DiGraph. The result represents a network in a distributed simulation.

    The Network classes (:class:`Network` and :class:`BidirectionalNetwork`) extend the Graph class and provides additional functionality
    for managing nodes and network properties.

    :param environment: The environment in which the network operates. If not provided, a new Environment instance will be created.
    :type environment: Environment, optional
    :param graph: The graph representing the network topology. Defaults to None.
    :type graph: NetworkX graph, optional
    :param kwargs: Additional keyword arguments.
    """

    def __init__(
        self,
        incoming_graph_data=None,  # Correct subclassing of Graph
        environment: Environment | None = None,
        behavioral_properties: NetworkBehaviorModel | None = None,
        **kwargs,
    ):
        super().__init__(incoming_graph_data)
        self._environment = environment or Environment2D()
        self.pos = {}
        self.ori = {}
        self.labels = {}
        self.simulation = None
        self.behavioral_properties = behavioral_properties or NetworkBehaviorModel.UnorderedCommunication
        logger.trace("Instance of Network has been initialized.")

    #### Overriding methods from Graph and DiGraph ####

    @staticmethod
    def to_directed_class():
        return Network

    @staticmethod
    def to_undirected_class():
        return BidirectionalNetwork

    def copy(self, as_view=False):
        """
        Return a copy of the graph. Parameters as_view is not used and is only for compatibility with NetworkX.

        :param as_view: Unused parameter. Defaults to False.
        :type as_view: bool, optional
        :return: A deepcopy of the graph.
        :rtype: NetworkMixin
        """
        return deepcopy(self)

    def __deepcopy__(self, memo, nodes=None, edges=None, init_args=None, cls=None):
        """
        Deepcopy the network.

        :param memo: A dictionary that maps object ids to the copied objects.
        :type memo: dict
        :param nodes: A list of nodes to include in the deepcopy. Defaults to None.
        :type nodes: Iterable[Node], optional
        :param edges: A list of edges to include in the deepcopy. Defaults to None.
        :type edges: Iterable[tuple[Node, Node]], optional
        :param init_args: A dictionary of additional arguments to pass to the constructor of the copied network. Defaults to None.
        :type init_args: dict, optional
        :param cls: The class to use for the copied network. Defaults to None.
        :type cls: type, optional
        """
        if id(self) in memo:
            return memo[id(self)]

        new_network = (cls or type(self))(**(init_args or {}))
        memo[id(self)] = new_network

        # Immutable attributes
        new_network._environment = self._environment
        new_network.behavioral_properties = self.behavioral_properties

        nodes = nodes or self.nodes()
        edges = edges or self.edges()

        nodes_copy = {node: deepcopy(node, memo) for node in nodes}
        for node, copy_node in nodes_copy.items():
            copy_node.network = None
            new_network.add_node(
                copy_node,
                pos=deepcopy(self.pos[node], memo),
                ori=deepcopy(self.ori[node], memo),
            )

        for u, v in edges:
            if u not in nodes or v not in nodes:
                continue
            new_network.add_edge(nodes_copy[u], nodes_copy[v])
        new_network.simulation = deepcopy(self.simulation, memo)

        new_network.clear_observers()
        return new_network

    def remove_node(self, node: "Node", skip_check=False):
        """
        Remove a node from the network.

        :param node: The node to be removed.
        :type node: Node
        :raises NetworkException: If the node is not in the network.
        """
        if not skip_check and node not in self.nodes():
            logger.error("Node not in network")
            raise NetworkException(NetworkException.ERRORS.NODE_NOT_IN_NET)
        super().remove_node(node)
        del self.pos[node]
        del self.labels[node]
        node.network = None
        logger.debug("Node with id {} is removed.", node._internal_id)

    def add_node(self, node=None, pos=None, ori=None, commRange=None):
        """
        Add node to network.

        :param node: node to add, default: new node is created
        :type node: Node, optional
        :param pos: position (x,y), default: random free position in environment
        :type pos: tuple, optional
        :param ori: orientation from 0 to 2*pi, default: random orientation
        :type ori: float, optional
        :param commRange: communication range of the node, default: None
        :type commRange: float, optional

        :return: the added node
        :rtype: Node

        :raises NetworkException: if the given node is already in another network or the given position is not free space

        """
        if not node:
            node = Node(commRange=commRange)

        assert isinstance(node, Node)

        if node.network:
            logger.exception("Node is already in another network, can't add.")
            raise NetworkException(NetworkException.ERRORS.NODE)

        node.network = self

        pos = pos if pos is not None else self._environment.find_random_pos(n=100)
        ori = ori if ori is not None else rand() * 2 * pi
        while ori > 2 * pi:
            ori -= 2 * pi

        if not self._environment.is_space(pos):
            logger.error("Given position is not free space.")
            raise NetworkException(NetworkException.ERRORS.NODE_SPACE)

        super().add_node(node)
        self.pos[node] = array(pos)
        self.ori[node] = ori
        self.labels[node] = str(node._internal_id)
        logger.debug("Node {} is placed on position {}.", node._internal_id, pos)
        self._copy_observers_to_nodes(node)
        return node

    def nodes(self, *args, **kwargs) -> Iterable["Node"]:
        return super().nodes(*args, **kwargs)

    #### Network attribute management ####

    @property
    def environment(self):
        return self._environment

    def _set_environment(self, environment: Environment):
        """
        Setter for environment. Override this method if default
        behavior is not enough.

        :param environment: The new environment for the network.
        :type environment: Environment
        """
        self._environment = environment

    @environment.setter
    def environment(self, environment: Environment):
        self._set_environment(environment)

    def add_observers(self, *observers):
        super().add_observers(*observers)
        self._copy_observers_to_nodes()

    def _copy_observers_to_nodes(self, *nodes):
        if not nodes:
            nodes = self.nodes()

        for node in nodes:
            node.add_observers(
                *(obs for obs in self.observers if isinstance(obs, NodeObserver))
            )  # Register observers to the nodes, too (only if they are NodeNetworkObservers)

    def clear_observers(self):
        super().clear_observers()
        for node in self.nodes():
            node.clear_observers()

    def reset(self, log=True):
        """
        Reset the network to its initial state.

        Does not reset the observers of the network nor the observers of the nodes.
        """
        logger.debug("Resetting network.")
        self.reset_all_nodes()

    def reset_all_nodes(self):
        """
        Reset all nodes in the network.

        :return: None
        """
        logger.debug("Resetting all nodes.")
        for node in self.nodes():
            node.reset()

            for other_node in self.out_neighbors(node):
                if node != other_node:
                    self.get_transit_messages(node, other_node).clear()
                    self.get_lost_messages(node, other_node).clear()

    #### Network methods ####

    def is_connected(self):
        if self.is_directed():
            return is_strongly_connected(self)
        return is_connected(self)

    def subnetwork(
        self,
        nodes: list[Node],
        edges: None | Iterable[tuple[Node, Node]] = None,
    ):
        """
        Returns a Network instance with the specified nodes and edges.

        :param nbunch: A list of nodes to include in the subnetwork.
        :type nbunch: list[Node]
        :param edges: A list of edges to include in the subnetwork. Defaults to None.
        :type edges: Iterable[tuple[Node, Node]], optional
        :return: A Network instance representing the subnetwork.
        :rtype: Network
        """

        return self.__deepcopy__({}, nodes, edges)

    def get_tree_net(self, tree_key, return_subnetwork=True):
        """
        Returns a new network with edges that are not in a tree removed.

        The tree is defined in the nodes' memory under the specified tree_key. The method iterates over all nodes in the network and checks if the tree_key is present in their memory. If it is, it retrieves the tree neighbors or children and adds the corresponding edges to the tree_edges_ids list. It also adds the tree nodes to the tree_nodes list.

        After iterating over all nodes, a subnetwork is created using the tree_nodes. Then, the method removes any edges from the subnetwork that are not present in the tree_edges_ids list.

        Finally, the resulting subnetwork, representing the tree, is returned.

        :param tree_key: The key in the nodes' memory that defines the tree. It can be a list of tree neighbors or a dictionary with 'parent' (node) and 'children' (list) keys.
        :type tree_key: str
        :return: A new network object with only the edges that are part of the tree.
        :param return_subnetwork: Whether to return a new network or the sets of nodes and edges. Defaults to True.
        :type return_subnetwork: bool, optional
        :rtype: NetworkMixin
        """
        tree_edges = []
        tree_nodes = []
        for node in self.nodes():
            neighbors_in_tree = []
            if tree_key not in node.memory:
                continue
            if isinstance(node.memory[tree_key], list):
                neighbors_in_tree = node.memory[tree_key]
            elif isinstance(node.memory[tree_key], dict):
                if "children" in node.memory[tree_key]:
                    neighbors_in_tree += [n.unbox() for n in node.memory[tree_key]["children"]]
                if "parent" in node.memory[tree_key] and node.memory[tree_key]["parent"] is not None:
                    neighbors_in_tree.append(node.memory[tree_key]["parent"].unbox())

            tree_edges.extend([(node, neighbor) for neighbor in neighbors_in_tree if neighbor in self.neighbors(node)])
            tree_nodes.extend(neighbor for neighbor in neighbors_in_tree if neighbor in self.neighbors(node))
            if neighbors_in_tree:
                tree_nodes.append(node)

        if not return_subnetwork:
            return tree_nodes, tree_edges

        treeNet = self.subnetwork(tree_nodes, tree_edges)

        return treeNet

    def nodes_sorted(self) -> list["Node"]:
        """
        Returned sorted nodes by id.

        :return: A sorted tuple of nodes.
        :rtype: list[Node]
        """
        return sorted(self.nodes(), key=lambda k: k._internal_id)

    def node_by_id(self, id_):
        """
        Returns the first node with the given id.

        :param id_: The id of the node to search for.
        :type id_: int
        :return: The node with the given id.
        :rtype: Node
        :raises NetworkException: If the network does not have a node with the given id.
        """
        for n in self.nodes():
            if n._internal_id == id_:
                return n

        logger.error("Network has no node with id {}.", id_)
        raise NetworkException(NetworkException.ERRORS.NODE_NOT_IN_NET)

    def avg_degree(self):
        """
        Calculate the average degree of the network.
        Uses out_degree for directed networks (amount of outgoing edges) and degree for undirected networks.

        :return: The average degree of the network.
        :rtype: float
        """
        degree_iter = self.out_degree() if self.is_directed() else self.degree()
        return average(tuple(map(lambda x: x[1], tuple(degree_iter))))

    def out_neighbors(self, node: "Node") -> set["Node"]:
        if self.is_directed():
            return set(self.successors(node))
        return set(self.neighbors(node))

    def in_neighbors(self, node: "Node") -> set["Node"]:
        if self.is_directed():
            return set(self.predecessors(node))
        return set(self.neighbors(node))

    def increment_node_clocks(self):
        """
        Increment the clock of all nodes in the network.

        Follows the clock increment function defined in the behavioral properties.

        :return: None
        """
        for node in self.nodes():
            node.clock += self.behavioral_properties._get_clock_increment(node)

    #### Node communication methods ####

    def get_transit_messages(self, u: "Node", v: "Node") -> dict["Message", int]:
        """
        Get messages in transit from node u to node v.

        :param u: The source node.
        :type u: Node
        :param v: The destination node.
        :type v: Node
        :return: A dictionary of messages in transit from node u to node v,
                 with the message as the key and the delay as the value.
        :rtype: Dict[Message, int]
        """
        if "in_transit_messages" not in self[u][v]:
            self[u][v]["in_transit_messages"] = {}
        return self[u][v]["in_transit_messages"]

    def add_transit_message(self, u: "Node", v: "Node", message: "Message", delay: int):
        """
        Add a message to the in-transit messages from node u to node v.

        :param u: The source node.
        :type u: Node
        :param v: The destination node.
        :type v: Node
        :param message: The message
        :type message: Message
        :param delay: The delay of the message
        :type delay: int
        """

        in_transit_messages = self.get_transit_messages(u, v)
        in_transit_messages[message] = delay

    def get_lost_messages(self, u: "Node", v: "Node") -> list["Message"]:
        """
        Get lost messages from node u to node v.

        :param u: The source node.
        :type u: Node
        :param v: The destination node.
        :type v: Node
        :return: A list of lost messages from node u to node v.
        :rtype: list[Message]
        """
        if "lost_messages" not in self[u][v]:
            self[u][v]["lost_messages"] = []
        return self[u][v]["lost_messages"]

    def add_lost_message(self, u: "Node", v: "Node", message: "Message"):
        """
        Add a message to the lost messages from node u to node v.

        :param u: The source node.
        :type u: Node
        :param v: The destination node.
        :type v: Node
        :param message: The message
        :type message: Message
        """
        self.get_lost_messages(u, v).append(message)

    def communicate(self):
        """
        Pass all messages from node's outboxes to its neighbors inboxes.

        This method collects messages from each node's outbox and delivers them to the appropriate destination.
        If the message is a broadcast, it is sent to all neighbors.
        If the message has a specific next hop, it is sent directly to that node.
        If the message has a specific destination, it is sent to that neighbor if it is reachable.

        :return: None
        """
        logger.debug("Communicating messages in the network.")

        if len(self) == 0:
            return

        transmission_complete: list["Message"] = []

        # Process messages in outboxes
        for node in self.nodes():
            # reversed to process messages in the order they were added
            for message in reversed(node.outbox.copy()):
                next_dest: "Node" = message.destination
                node.outbox.remove(message)

                if self.behavioral_properties._should_lose(self, message):
                    logger.trace("Message lost: {}", message)
                    self.add_lost_message(node, next_dest, message)
                    continue

                self.add_transit_message(
                    node,
                    next_dest,
                    message,
                    self.behavioral_properties._get_delay(self, message),
                )

        # Process messages in transit
        for u, v in self.edges():
            messages_delay = self.get_transit_messages(u, v)
            if not messages_delay:
                continue

            message_ordering = self.behavioral_properties.message_ordering

            # Decrease delay for all messages
            for message in messages_delay:
                messages_delay[message] -= 1

            # Sort messages by id
            messages = sorted(messages_delay.keys(), key=lambda x: x._internal_id)
            if not message_ordering:
                # Artificially create inversions in the message order
                messages = sorted(messages_delay.keys(), key=lambda x: x._internal_id)
                cant_inversions = len(messages) // 4
                artificial_inversions = choices(
                    [(i, j) for i in range(len(messages)) for j in range(len(messages)) if i < j],
                    k=cant_inversions,
                )

                for i, j in artificial_inversions:
                    messages[i], messages[j] = messages[j], messages[i]

            for message in messages:
                if messages_delay[message] > 0 and message_ordering:
                    # When a delayed message is found, only process the rest if there is no message ordering.
                    # This is to ensure that the messages are processed in order, even if a previous message is delayed.
                    break
                if messages_delay[message] <= 0:
                    transmission_complete.append(message)
                    messages_delay.pop(message)

        for message in transmission_complete:
            logger.debug("Communicating message: {}", message)

            if message.destination is not None:
                # Destination is neighbor
                if message.source in self and message.destination in self.neighbors(
                    message.source
                ):  # for DiGraph, `self.neighbors` are the out-neighbors
                    self.deliver_to(message.destination, message)
                else:
                    raise MessageUndeliverableException("Can't deliver message.", message)

    def deliver_to(self, destination: "Node", message: "Message"):
        """
        Deliver a message to a destination node in the network.

        :param destination: The destination node to send the message to.
        :type destination: Node
        :param message: The message to be sent.
        :type message: Message
        :raises PyDistSimMessageUndeliverable: If the destination is not in the network.
        """
        logger.debug("Sending message from {} to {}.", message.source, destination)
        if destination in self.nodes():
            destination.push_to_inbox(message)
        else:
            raise MessageUndeliverableException("Destination not in network.", message)

    #### Visualization methods ####

    def show(self, *args, **kwargs):
        """
        Get and show the figure object representing the network visualization.

        :param args: Additional positional arguments to pass to the draw function.
        :param kwargs: Additional keyword arguments to pass to the draw function.
        :return: The figure object representing the network visualization.
        :rtype: matplotlib.figure.Figure
        """
        fig = self.get_fig(*args, **kwargs)
        fig.show()
        return fig

    def savefig(self, fname="network.png", figkwargs={}, *args, **kwargs):
        self.get_fig(*args, **kwargs).savefig(fname, **figkwargs)

    def get_fig(self, *args, **kwargs):
        """
        Get the figure object representing the network visualization.

        :param args: Additional positional arguments to pass to the draw function.
        :param kwargs: Additional keyword arguments to pass to the draw function.
        :return: The figure object representing the network visualization.
        :rtype: matplotlib.figure.Figure
        """
        return draw.draw_current_state(
            self,
            *args,
            **kwargs,
        )

    def get_size(self):
        """Returns network width and height based on nodes positions."""
        return max(list(self.pos.values()), axis=0) - min(list(self.pos.values()), axis=0)

    def get_dic(self):
        """
        Return all network data in the form of a dictionary.

        :return: A dictionary containing the network data.
        :rtype: dict
        """
        pos = {
            n: f"x: {self.pos[n][0]:.2f} y: {self.pos[n][1]:.2f} theta: {self.ori[n] * 180.0 / pi:.2f} deg"
            for n in self.nodes_sorted()
        }
        edges = [(x_node._internal_id, y_node._internal_id) for x_node, y_node in self.edges()]
        return {
            "nodes": pos,  # A dictionary mapping nodes to their positions in the network.
            "edges": edges,  # A list of pairs (id1, id2) representing the edges by node id.
        }

    #### Test helper methods ####

    def validate_params(self, params: dict):
        """
        Validate if given network params match its real params.

        :param params: A dictionary containing the network parameters to validate.
        :type params: dict

        :raises AssertionError: If any of the network parameters do not match the real parameters.
        """
        logger.info("Validating params...")
        count = params.get("count", None)  #  for unit tests
        if count:
            if isinstance(count, list):
                assert len(self) in count, f"{len(self)=} not in {count=}"
            else:
                assert len(self) == count, f"{len(self)=} != {count=}"
        n_min = params.get("n_min", 0)
        n_max = params.get("n_max", Inf)
        assert len(self) >= n_min and len(self) <= n_max
        for param, value in params.items():
            if param == "connected":
                assert not value or self.is_connected(), f"{value=}, {self.is_connected()=}"
            elif param == "degree":
                assert allclose(self.avg_degree(), value, atol=1)
            elif param == "environment":
                assert self.environment.__class__ == value.__class__
            elif param == "sensors":
                compositeSensor = CompositeSensor(Node(), value)
                for node in self:
                    assert all(
                        map(
                            lambda s1, s2: pydistsim_equal_objects(s1, s2),
                            node.sensors,
                            compositeSensor.sensors,
                        )
                    )
            elif param == "aoa_pf_scale":
                for node in self:
                    for sensor in node.sensors:
                        if sensor.name() == "AoASensor":
                            assert sensor.probabilityFunction.scale == value
            elif param == "dist_pf_scale":
                for node in self:
                    for sensor in node.sensors:
                        if sensor.name() == "DistSensor":
                            assert sensor.probabilityFunction.scale == value


class Network(NetworkMixin, DiGraph):
    """
    A directed graph representing a network in a distributed simulation.
    """


class BidirectionalNetwork(NetworkMixin, Graph):
    """
    An undirected graph representing a bidirectional network in a distributed simulation.
    """


NetworkType = Network | BidirectionalNetwork
"A network in a distributed simulation."



"""
Base classes for restrictions over message passing networks.

All restrictions are based on "Design and Analysis of Distributed Algorithms" by Nicola Santoro.

There are two types of restrictions: checkable and applicable. Checkable restrictions can be checked for a network,
while applicable restrictions can be applied to a network. Sometimes, an applicable restriction may check itself.
"""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pydistsim.network.network import NetworkType


def abstractclassmethod(method):
    return classmethod(abstractmethod(method))


class Restriction(ABC):
    """
    Base class for all restrictions over a message passing network.
    """


class CheckableRestriction(Restriction, ABC):
    """
    Base class for checkable restrictions over a message passing network.
    """

    @abstractclassmethod
    def check(cls, network: "NetworkType") -> bool: ...

    @classmethod
    def get_help_message(cls, network: "NetworkType") -> str:
        return cls.help_message


class ApplicableRestriction(Restriction, ABC):
    """
    Base class for restrictions that can be applied to a network.

    This is a separate class from :class:`Restriction` to allow for restrictions that are not checkable.
    """

    help_message: str = (
        "The easiest fix would be adding the line `self.apply_restrictions()` to the initializer method of the "
        "algorithm."
    )

    @abstractclassmethod
    def apply(self, network: "NetworkType") -> None: ...



from pydistsim.benchmark import AlgorithmBenchmark, MetricCollector
from pydistsim.demo_algorithms.broadcast import Flood
from pydistsim.demo_algorithms.santoro2007.yoyo import YoYo
from pydistsim.network.behavior import NetworkBehaviorModel
from pydistsim.utils.testing import PyDistSimTestCase


class MyCustomMetricCollector(MetricCollector):

    def create_report(self):
        report = super().create_report()

        report["test_qty"] = 42
        report["event_qty"] = len(self.metrics)

        return report


class TestBenchmark(PyDistSimTestCase):
    def test_flood_benchmark(self):
        benchmark = AlgorithmBenchmark(
            ((Flood, {"initial_information": "Hello Wold Test!"}),),
            network_sizes=range(1, 15),
            network_behavior=NetworkBehaviorModel.UnorderedRandomDelayCommunication,
            metric_collector_factory=MyCustomMetricCollector,
            max_time=60,
        )

        benchmark.run()

        df = benchmark.get_results_dataframe()

        assert "test_qty" in df.columns and "event_qty" in df.columns

        benchmark.plot_analysis(
            x_vars=["Network type"],
            y_vars=["event_qty"],
            result_filter=lambda df: df["Network type"] in ("complete", "ring"),
            grouped=False,
        )

        benchmark.plot_analysis()

    def test_yoyo_benchmark(self):
        yoyo_benchmark = AlgorithmBenchmark(
            (YoYo,),
            network_sizes=range(1, 15),
            metric_collector_factory=MyCustomMetricCollector,
            max_time=60,
        )

        yoyo_benchmark.run()

        df = yoyo_benchmark.get_results_dataframe()

        assert "test_qty" in df.columns and "event_qty" in df.columns

        yoyo_benchmark.plot_analysis(
            x_vars=["Network type"],
            y_vars=["event_qty"],
            result_filter=lambda df: df["Network type"] in ("complete", "ring"),
            grouped=False,
        )

        yoyo_benchmark.plot_analysis()
